# This file is a part of ducktools.pytui
# A TUI for managing Python installs and virtual environments
#
# Copyright (C) 2025  David C Ellis
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path
import typing

from ducktools.pythonfinder import list_python_installs, PythonInstall
from ducktools.pythonfinder.venv import get_python_venvs, PythonVEnv

from textual.app import App
from textual.binding import Binding
from textual.containers import Vertical
from textual.widgets import DataTable, Footer, Header, Label

from .commands import launch_repl, launch_shell, create_venv


DATATABLE_BINDINGS_NO_ENTER = [b for b in DataTable.BINDINGS if b.key != "enter"]
CWD = Path.cwd()


class VEnvTable(DataTable):
    BINDINGS = [
        Binding(key="enter", action="app.activated_shell", description="Activate VEnv and Launch Shell", show=True),
        Binding(key="r", action="app.launch_venv_repl", description="Launch VEnv Python REPL", show=True),
        # Binding(key="p", action="show_packages", description="List Installed Packages", show=True),
        *DATATABLE_BINDINGS_NO_ENTER,
    ]

    def on_mount(self):
        self.load_venvs()

    def load_venvs(self):
        self.loading = True
        try:
            self.cursor_type = "row"
            self.add_columns("Version", "Environment Path", "Runtime Path")
            for venv in get_python_venvs(base_dir=CWD, recursive=False, search_parent_folders=True):
                folder = str(Path(venv.folder).relative_to(CWD))
                self.add_row(venv.version_str, folder, venv.parent_path, key=venv)
        finally:
            self.loading = False


class RuntimeTable(DataTable):
    BINDINGS = [
        # Binding(key="v", action="app.create_venv", description="Create Virtual Environment", show=True),
        Binding(key="r", action="app.launch_runtime", description="Launch Runtime Python REPL", show=True),
        *DATATABLE_BINDINGS_NO_ENTER
    ]

    def on_mount(self):
        self.load_runtimes()

    def load_runtimes(self):
        self.loading = True
        try:
            self.cursor_type = "row"
            self.add_columns("Version", "Managed By", "Implementation", "Path")
            for install in list_python_installs():
                path = str(Path(install.executable).parent)
                self.add_row(
                    install.version_str,
                    install.managed_by,
                    install.implementation,
                    path,
                    key=install
                )
        finally:
            self.loading = False


class ManagerApp(App):
    BINDINGS = [
        Binding(key="q", action="quit", description="Quit"),
    ]

    CSS = """
    .boxed {
        height: auto;
        border: solid green;
    }
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._venv_table = VEnvTable()
        self._runtime_table = RuntimeTable()

    def compose(self):
        yield Header()
        with Vertical(classes="boxed"):
            yield Label("Virtual Environments")
            yield self._venv_table
        with Vertical(classes="boxed"):
            yield Label("Python Runtimes")
            yield self._runtime_table
        yield Footer()

    def action_launch_runtime(self):
        table = self._runtime_table

        row = table.coordinate_to_cell_key(table.cursor_coordinate)

        install = typing.cast(PythonInstall, row.row_key.value)
        python_exe = install.executable

        # Suspend the app and launch python
        # Ignore keyboard interrupts otherwise the program will exit when this exits.
        with self.suspend():
            launch_repl(python_exe)

        # Redraw
        self.refresh()

    def action_launch_venv_repl(self):
        table = self._venv_table
        row = table.coordinate_to_cell_key(table.cursor_coordinate)

        install = typing.cast(PythonVEnv, row.row_key.value)
        python_exe = install.executable

        # Suspend the app and launch python
        # Ignore keyboard interrupts otherwise the program will exit when this exits.
        with self.suspend():
            launch_repl(python_exe)

        # Redraw
        self.refresh()

    def action_activated_shell(self):
        table = self._venv_table
        row = table.coordinate_to_cell_key(table.cursor_coordinate)
        venv = typing.cast(PythonVEnv, row.row_key.value)
        with self.suspend():
            launch_shell(venv)

        # Redraw
        self.refresh()

    def on_mount(self):
        self.title = "Ducktools.PyTui: Python Environment and Runtime Manager"



# This file is a part of ducktools.pytui
# A TUI for managing Python installs and virtual environments
#
# Copyright (C) 2025  David C Ellis
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
from pathlib import Path

import shellingham
from ducktools.pythonfinder.venv import PythonVEnv

from .util import run


def launch_repl(python_exe: str) -> None:
    run([python_exe])


def create_venv(python_exe: str, venv_path: str = ".venv", include_pip: bool = True) -> Path:
    # Unlike the regular venv command defaults this will create an environment
    # and download the *newest* pip (assuming the parent venv includes pip)

    if include_pip:
        run([python_exe, "-m", "venv", "--upgrade-deps", venv_path])
    else:
        run([python_exe, "-m", "venv", "--without-pip", venv_path])

    if sys.platform == "win32":
        python_path = Path(venv_path).resolve() / "Scripts" / "python.exe"
    else:
        python_path = Path(venv_path).resolve() / "bin" / "python"

    return python_path


def launch_shell(venv: PythonVEnv) -> None:
    # Launch a shell with a virtual environment activated.
    env = os.environ.copy()
    old_path = env.get("PATH", "")
    env["PATH"] = os.pathsep.join([str(Path(venv.executable).parent), old_path])
    env["VIRTUAL_ENV"] = venv.folder
    env["VIRTUAL_ENV_PROMPT"] = Path(venv.folder).name

    for t, v in env.items():
        if type(v) is not str:
            assert False, t

    try:
        shell_name, shell = shellingham.detect_shell()
    except shellingham.ShellDetectionFailure:
        if os.name == "posix":
            shell_name, shell = "UNKNOWN", os.environ["SHELL"]
        elif os.name == "nt":
            shell_name, shell = "UNKNOWN", os.environ["COMSPEC"]
        else:
            raise RuntimeError(f"Shell detection failed")

    if shell_name == "cmd":
        old_prompt = env.get("PROMPT", "$P$G")
        env["PROMPT"] = f"({Path(venv.folder).name}) {old_prompt}"
    elif shell_name == "bash":
        old_prompt = env.get("PS1", "")
        env["PS1"] = f"({Path(venv.folder).name}) {old_prompt}"

    print(f"Launching Shell with active VENV: {venv.folder}")
    print("Type 'exit' to close")
    run([shell], env=env)



# This file is a part of ducktools.pytui
# A TUI for managing Python installs and virtual environments
#
# Copyright (C) 2025  David C Ellis
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# This is a generic utility wrapper to launch a subprocess
# while ignoring specific signals in the parent process
import functools
import signal
import subprocess


class IgnoreSignals:
    @staticmethod
    def null_handler(signum, frame):
        # This just ignores signals, used to ignore in the parent process temporarily
        # The child process will still receive the signals.
        pass

    def __init__(self, signums: list[int]):
        self.old_signals = {}
        self.signums = signums

    def __enter__(self):
        if self.old_signals:
            raise RuntimeError("ignore_signals is not reentrant")

        for signum in self.signums:
            self.old_signals[signum] = signal.signal(signum, self.null_handler)

    def __exit__(self, exc_type, exc_val, exc_tb):
        for signum, handler in self.old_signals.items():
            signal.signal(signum, handler)


def ignore_keyboardinterrupt():
    return IgnoreSignals([signal.SIGINT])


@functools.wraps(subprocess.run, assigned=("__doc__", "__type_params__", "__annotations__"))
def run(*args, **kwargs):
    with ignore_keyboardinterrupt():
        subprocess.run(*args, **kwargs)
